from __future__ import annotations
# Methods to render the DataFrame chain as a set of expressions.
import ast
from typing import Dict, Optional, Union, Tuple

from dataframe_expressions import Column, DataFrame, ast_Callable, ast_Column, ast_DataFrame
from .utils_ast import CloningNodeTransformer


class ast_Filter (ast.AST):
    '''
    Represents a filter - the previous expression, which should be a
    sequence, needs to be filtered on this
    '''

    _fields = ('expr', 'filter')

    def __init__(self, expr: ast.AST, filter: ast.AST):
        '''
        Create a filter expression.

        Arguments:
            expr        AST of the filter expression
            filter      An expression that evaluates to bool, to be applied
                        on each item of the `expr`'s sequence.
        '''
        self.expr = expr
        self.filter = filter


class render_context:
    '''
    Class for internal use - maintains context and references as we move
    through the resolution. While this is returned to user code, it should
    not be accessed by user code!
    '''
    def __init__(self, template: Optional[render_context] = None):
        if template is None:
            self._seen_datasources: Dict[int, ast_DataFrame] = {}
            self._resolved: Dict[int, ast.AST] = {}
        else:
            self._seen_datasources = template._seen_datasources.copy()
            self._resolved = template._resolved.copy()

    def _lookup_dataframe(self, df: DataFrame) -> ast_DataFrame:
        '''
        See if a raw dataframe has already been tagged. If so, make sure
        we return the same object.
        '''
        h = hash(str(df))
        if h not in self._seen_datasources:
            self._seen_datasources[h] = ast_DataFrame(df)
        return self._seen_datasources[h]

    def _resolve_ast(self, a: ast.AST) -> ast.AST:
        '''
        Look to see if this `ast.AST` has already been run, and if so, return
        the same object to make downstream processing (and loop connection)
        easier.
        '''
        h = hash(ast.dump(a))
        if h not in self._resolved:
            self._resolved[h] = a
        return self._resolved[h]


class _parent_subs(CloningNodeTransformer):
    @classmethod
    def transform(cls, a: ast.AST,
                  context: render_context) -> ast.AST:
        v = _parent_subs(context)
        return v.visit(a)

    def __init__(self,
                 context: render_context):
        ast.NodeTransformer.__init__(self)
        self._context = context

    def visit_ast_Column(self, a: ast_Column):
        'We have a column embedded here. Sort it out'
        return _render_filter(a.column, self._context)

    def visit_ast_DataFrame(self, a: ast_DataFrame):
        'Sort out an embedded column'
        expr = render(a.dataframe, self._context)[0]
        return expr


def _render_filter(f: Column, context: render_context) \
        -> ast.AST:
    'Render a filter/Mask as a result'
    v = _parent_subs.transform(f.child_expr, context)
    assert v is not None
    return v


def render(d: Union[DataFrame, Column], in_context: Optional[render_context] = None) \
        -> Tuple[ast.AST, render_context]:
    '''
    Follows the data frame back to the start and renders it in a complete AST.

    Arguments:
        d           DataFrame rendered from the start

    Returns:
        expr        A python ast.AST that contains the complete expression for whatever this
                    dDataFrame is meant to represent. It will include the special `ast_Filter`
                    nodes that indicate a filtering operation is to take place

    Notes:
        In many cases, expressions are repeated. For example, `df[(df.x > 10) & (df.y > 10)]`,
        implies iterating over df. The `ast.AST` that represents `df` will be the same object
        in this case. That means the object hash will be the same. This can be used as a
        poor-person's way of doing common sub-expression elimination.
    '''
    context = render_context() if in_context is None else in_context

    # Simple out
    if isinstance(d, DataFrame) and d.child_expr is None:
        return context._lookup_dataframe(d), context

    # If this is a column, then it is a comparison expression.
    if isinstance(d, Column):
        return _render_filter(d, context), context

    # now we need to tack on our info.
    assert d.child_expr is not None
    expr = _parent_subs.transform(d.child_expr, context)
    if d.filter is not None:
        filter_expr = _render_filter(d.filter, context)
        expr = ast_Filter(expr, filter_expr)

    return context._resolve_ast(expr), context


def render_callable(callable: ast_Callable, context: render_context, *args) \
        -> Tuple[ast.AST, render_context]:
    '''
    A callable is invoked with the given list of arguments.

    Arguments:
        callable            The parsed out function all (lambda, or a function proper)
        context             The context to use when parsing. Will not be touched or updated.
        args                List of positional arguments to be passed to the lambda. They can
                            be any type, including data frame arguments.

    Returns:
        expr                A python `ast.AST` that contains the complete expression for whatever
                            this function returns. The expression follows the same rules as the
                            return for the `render` function.
        context             New context which are things already seen plus anything new.
    '''
    # Invoke the call
    d_result = callable.callable(*args)
    new_context = render_context(context)

    # Render it
    if isinstance(d_result, DataFrame):
        return render(d_result, new_context)[0], new_context
    elif isinstance(d_result, Column):
        return _render_filter(d_result, new_context), new_context
    else:
        from .utils import _term_to_ast
        return _term_to_ast(d_result, DataFrame()), new_context



from __future__ import annotations
import ast
from dataframe_expressions.asts import ast_DataFrame
from typing import Callable, List, Optional, cast
from contextlib import contextmanager

from .DataFrame import DataFrame


def _reset_alias_catalog():
    'For testing - reset to ground state'
    global _alias_catalog
    _alias_catalog = {}


class _alias_info:
    '''
    Hold onto all info we need to resolve an alias
    '''

    _inuse_stack: List[_alias_info] = []

    def __init__(self, base_obj: str, name: str, func_define: Callable[[DataFrame], DataFrame]):
        self._base_obj = base_obj
        self._func = func_define
        self._name = name

    def apply(self, df: DataFrame) -> DataFrame:
        @contextmanager
        def keep_queue_clean(a: _alias_info):
            _alias_info._inuse_stack.append(a)
            try:
                yield a
            finally:
                _alias_info._inuse_stack.pop()

        with keep_queue_clean(self):
            return self._func(df)


# List of the known aliases
_alias_catalog = {}


def define_alias(base_obj: str, name: str, func_define: Callable[[DataFrame], DataFrame]):
    '''
    Define an alias that can be used as a short cut for a reference later on. For example:

        ```
        df = DataFrame()
        define_alias (".jets", "pts", lambda o: o.pt/1000.0)
        ```

    When you write `df.jets.pts` it is the same as writing `df.jets.pt/1000.0`.

    If you've already used an alias in constructing a `DataFrame` then it will not be
    filled: alias are resolved as the DAG is being constructed, not when it is rendered.

    ## Adding aliases of the same name

    This implementation does no checking as things are added. You can have different base names,
    and if the base_obj's are different they will match differently, as you would expect.
    But they are evaluated last one entered first, so if you enter a complete wildcard, nothing
    else will ever get looked at. If this wasn't a prototype perhaps I'd not allow this.

    ## How `base_obj` works

        - The leading `.` indicates a reference from the root dataframe
        - Without the leading `.` it will just match from the current position. For example,
        if `base_obj` is `""`, then anytime you write `xxx.pts` it will become `xxx.pt/1000.0`.\
        This allows some level of wildcard.

    ## Restrictions on `func_define`

        - This function is called with a DataFrame as its argument. You may only do what you
          would normally do. So not loops, etc. It must return a new DataFrame!! Not even
          constant returns are supported just yet!

    Arguments:
        base_obj            Base objects, starting with a `.` indicating path to where this should
                            be applied.
        name                Name of the new functions
        func_define         Lambda that returns the modified DAG. See restrictions above for rules
                            on this lambda.
    '''
    a_info = _alias_info(base_obj, name, func_define)
    if name in _alias_catalog:
        _alias_catalog[name].append(a_info)
    else:
        _alias_catalog[name] = [a_info]


def _matches_pattern_str(df: DataFrame, a: str) -> bool:
    '''
    Return true if this matches the pattern specificed in a
    '''
    # Simple cases first
    if a is None or len(a) == 0:
        return True

    # If we are at a "." then we just need the parent
    # to be another dataframe for the attribute.
    if a == '.':
        return df.child_expr is None

    # We don't do total expression replacement, sadly
    if not isinstance(df.child_expr, ast.Attribute):
        return False

    # Next, make sure this is an attribute down a level
    if not isinstance(df.child_expr.value, ast_DataFrame):
        return False

    # Next, go down a level.
    parts = a.split('.')
    name = parts[-1]
    if name != df.child_expr.attr:
        return False

    a_minus = '.'.join(parts[:-1])
    if len(parts) == 2 and parts[0] == '':
        a_minus = '.'
    df_parent = cast(ast_DataFrame, df.child_expr.value)
    return _matches_pattern_str(df_parent.dataframe, a_minus)


def _matches_pattern(df: DataFrame, a: _alias_info) -> bool:
    '''
    Return true if this matches the patter specified in a.
    '''
    return _matches_pattern_str(df, a._base_obj)


def lookup_alias(df: DataFrame, name: str) -> Optional[_alias_info]:
    if name not in _alias_catalog:
        return None

    # Now see if they pattern match
    for a in _alias_catalog[name]:
        if a not in _alias_info._inuse_stack:
            if _matches_pattern(df, a):
                return a

    return None



import ast
from typing import Callable, cast


class ast_DataFrame(ast.AST):
    '''
    Hold onto a dataframe reference
    '''
    _fields = ()

    def __init__(self, dataframe=None):
        ast.AST.__init__(self)
        from .DataFrame import DataFrame
        assert dataframe is None or isinstance(dataframe, DataFrame)
        self.dataframe = cast(DataFrame, dataframe)


class ast_Column(ast.AST):
    '''Ast that holds onto a DataFrame reference'''
    _fields = ()

    def __init__(self, col=None):
        ast.AST.__init__(self)
        from .DataFrame import Column
        assert col is None or isinstance(col, Column)
        self.column = cast(Column, col)


class ast_Callable(ast.AST):
    'An AST node that is some sort of python callable, along with the df it was called from.'

    _fields = ('name',)

    def __init__(self, callable: Callable = None, relative_to=None):
        '''
        relative_to is optional - in which case this is a function call, not an
        extension method!
        '''
        ast.AST.__init__(self)
        if callable is not None:
            if callable.__name__ == '<lambda>':
                self.name = f'lambda-{callable.__code__.co_filename}:' \
                            f'{callable.__code__.co_firstlineno}'
            else:
                self.name = callable.__name__
        self.callable = callable
        self.dataframe = relative_to


class ast_FunctionPlaceholder(ast.AST):
    'An AST node that represents a function to be called, that is a placeholder'

    _fields = ('name',)

    def __init__(self, callable: Callable = None):
        ast.AST.__init__(self)
        self.name = callable.__name__
        self.callable = callable



from .DataFrame import Column, DataFrame  # NOQA
from .alias import define_alias  # NOQA
from .asts import (  # NOQA
    ast_Callable, ast_Column, ast_DataFrame, ast_FunctionPlaceholder)
from .render import ast_Filter, render, render_callable, render_context  # NOQA
from .utils import DataFrameTypeError, user_func, exclusive_class  # NOQA


from __future__ import annotations
import ast
import logging
from typing import Any, Callable, Dict, List, Optional, Tuple, Union, cast

from .asts import ast_Callable, ast_DataFrame
from .utils_ast import CloningNodeTransformer


class Column:
    '''
    Base class for a single sequence. Unlike a `DataFrame` this can't have any complex structure.
    It is a sequence of items, assumed to be of the same type.
    '''
    def __init__(self, t: Any, expr: ast.AST):
        self.child_expr: ast.AST = expr
        self._fields = ('child_expr',)
        self.type = t

    def __and__(self, other) -> Column:
        ''' Bitwise and becomes a logical and. '''
        from .utils import _term_to_ast
        return Column(type(bool), ast.BoolOp(op=ast.And(),
                      values=[_term_to_ast(self, self), _term_to_ast(other, self)]))

    def __or__(self, other) -> Column:
        ''' Bitwise and becomes a logical and. '''
        from .utils import _term_to_ast
        return Column(type(bool), ast.BoolOp(op=ast.Or(),
                      values=[_term_to_ast(self, self), _term_to_ast(other, self)]))

    def __invert__(self) -> Column:
        ''' Invert, or logical NOT operation. '''
        from .utils import _term_to_ast
        return Column(type(bool), ast.UnaryOp(op=ast.Invert(),
                      operand=_term_to_ast(self, self)))


class _sub_link_info:
    '''
    Info on links between dataframes or functions that modify data frames
    '''
    def __init__(self, df: Union[DataFrame, Callable[[DataFrame], DataFrame]],
                 computed_col: bool):
        self._df = df
        self.computed_col = computed_col

    def render(self, df: DataFrame) -> DataFrame:
        '''
        Given the parent DF, render the substitution
        '''
        if isinstance(self._df, DataFrame):
            return self._df
        else:
            assert callable(self._df), 'Internal Error - bad substitution'
            r = ast_Callable(self._df, df)
            expr = ast.Call(func=r, args=[ast_DataFrame(df)])
            return DataFrame(expr=expr)


def _do_not_extend(o: object):
    'Test if the object has the "no-extend" flag'
    return '__no_arb_attr' in dir(o)


class DataFrame:
    '''
    Base class for building a dataframe expression.

    Notes:
        - Any properties we have here will hide the name of a column in this data frame
    '''
    def __init__(self,
                 expr: Optional[ast.AST] = None,
                 filter: Optional[Column] = None,
                 df_to_copy: Optional[DataFrame] = None):
        '''
        Create the base DataFrame that is at the top of the parse tree.

        Arguments
            p           Parent dataframe we are coming from
            expr        Expression to be applied to the parent
            filter      A filter to be applied
            df_to_copy  A dataframe to be copied from (like a clone operation)
        '''
        self.child_expr: Optional[ast.AST] = expr
        self.filter: Optional[Column] = filter

        self._sub_df: Dict[str, _sub_link_info] = {}

        if df_to_copy is not None:
            self._sub_df = df_to_copy._sub_df

    def check_attribute_name(self, name) -> None:
        'Throw an error if the attribute name is bad'
        pass

    def _find_compat_parent_attribute(self, name, computed_col_only: bool = True) \
            -> Optional[Tuple[DataFrame, DataFrame, List[Column]]]:
        '''
        Find a compatible parent's attribute. If not compatible, then
        return none.
        '''
        p = self
        filters: List[Column] = []
        while p is not None:
            if p.filter is not None:
                filters.append(p.filter)

            # Make sure we aren't changing the data - if so, then we
            # can't walk back up further.
            if not isinstance(p.child_expr, ast_DataFrame):
                return None

            p = p.child_expr.dataframe
            if name in p._sub_df and ((not computed_col_only) or p._sub_df[name].computed_col):
                # Defined column with extension mechanism
                expr = p._sub_df[name].render(p)
                return expr, p, filters

            if name in dir(p):
                # Column is defined in the object
                # We don't call hasattr as we don't want to generate a new attribute.
                expr = getattr(p, name)
                return expr, p, filters

        return None

    def _replace_root_expr(self, parent: DataFrame, filters: List[Column]):
        '''
        Look through our self, and anything attached to us for the parent dataframe.
        Once we've found it, create a new one, and attach all filters to it.
        '''
        if self is parent:
            # Ok - we found the parent dataframe. Time to create new dataframes with
            # all filters strung on it.
            df = self
            for f in filters:
                df = DataFrame(expr=ast_DataFrame(df), filter=f)
            return df

        # Now we need to recurse and find all data frames and rebuild them.
        class search_for_ast(CloningNodeTransformer):
            def __init__(self):
                CloningNodeTransformer.__init__(self)

            def visit_ast_DataFrame(self, a: ast_DataFrame):
                new_df = a.dataframe._replace_root_expr(parent, filters)
                if new_df is a.dataframe:
                    return a
                return ast_DataFrame(new_df)

        sa_transform = search_for_ast()
        new_child = None if self.child_expr is None else sa_transform.visit(self.child_expr)
        if new_child is not self.child_expr:
            return DataFrame(expr=new_child, filter=self.filter, df_to_copy=self)
        return self

    def __getattr__(self, name: str) -> DataFrame:
        '''Reference a column name'''
        if name.startswith('_'):
            raise AttributeError(name)

        # Have we done this before?
        if name not in self._sub_df:
            result = None

            # Resolve any aliases we need
            if result is None:
                from .alias import lookup_alias
                a = lookup_alias(self, name)
                if a is not None:
                    result = a.apply(self)

            # Is this attribute used by anyone above us?
            if result is None:
                p_attr = self._find_compat_parent_attribute(name)
                if p_attr is not None:
                    # Tricky part - we need any of the filters that were accumulated applied.
                    attr, parent, filters = p_attr
                    df = attr._replace_root_expr(parent, filters)
                    result = df

            # Ok - in that case, this is a straight attribute, as long
            # as we are allowed to do the lookup.
            if result is None:
                if _do_not_extend(self):
                    # Oooo - they are trying to access something we don't know about!
                    raise Exception(f'No such attribute explicitly defined ("{name}")')

                child_expr = ast.Attribute(value=ast_DataFrame(self), attr=name,
                                           ctx=ast.Load())
                result = DataFrame(expr=child_expr)

            self._sub_df[name] = _sub_link_info(result, False)
        return self._sub_df[name].render(self)

    def __getitem__(self, expr: Union[Callable, DataFrame, Column, int]) -> DataFrame:
        '''A filtering operation of some sort'''
        assert isinstance(expr, (DataFrame, Column, int)) or callable(expr), \
            "Filtering a data frame must be done by a DataFrame expression " \
            f"(type: DataFrame or Column or int) not '{type(expr).__name__}'"

        if isinstance(expr, int):
            c_expr = ast.Subscript(
                value=ast_DataFrame(self),
                slice=ast.Index(value=expr)
            )
            return DataFrame(expr=c_expr)

        if callable(expr) and not (isinstance(expr, DataFrame) or isinstance(expr, Column)):
            c_expr = expr(self)
            assert isinstance(c_expr, DataFrame) or isinstance(c_expr, Column), \
                f"Filter function '{expr.__name__}'' did not return a DataFrame expression"
            expr = c_expr

        if isinstance(expr, DataFrame):
            assert expr.filter is None
            assert expr.child_expr is not None
            expr = Column(bool, expr.child_expr)
        # Redundant, but above too complex for type processor?
        assert isinstance(expr, Column), 'Internal error - filter must be a bool column!'
        return DataFrame(ast_DataFrame(self), filter=expr)

    def __setitem__(self, key: str,
                    expr: Union[DataFrame, Callable[[DataFrame], DataFrame]]) \
            -> DataFrame:
        '''
        Add a new leaf to this data frame
        '''
        assert isinstance(key, str)
        assert len(key) > 0
        if key in self._sub_df:
            if not self._sub_df[key].computed_col:
                raise Exception(f'You may not redefine "{key}".')
            else:
                logging.getLogger(__name__).warning('')

        self._sub_df[key] = _sub_link_info(expr, True)
        return self

    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs) -> Any:
        '''Take over a numpy or similar execution by turning it into a function call'''
        visitor = getattr(self, ufunc.__name__, None)
        assert visitor is not None, f'Unable to call function "{ufunc.__name__}" on dataframe.'
        return visitor(*inputs[1:], **kwargs)

    def __call__(self, *inputs, **kwargs) -> DataFrame:
        '''
        Someone is trying to turn an attribute into a function. That is fine, but it takes some
        fancy footwork on our part. Specifically, what we were thinking of as an attribute is
        actual a function call. So we have to haul that back to undo the attribute and turn it
        into a function call.
        '''
        assert self.child_expr is not None, \
            'Cannot call a DataFrame directly - must be a function name!'
        assert isinstance(self.child_expr, ast.Attribute), \
            'Cannot call a DataFrame directly - must be a function name!'
        from .utils import _term_to_ast
        assert isinstance(self.child_expr, ast.Attribute)
        assert isinstance(self.child_expr.value, ast_DataFrame)
        base_df = cast(ast_DataFrame, self.child_expr.value)
        child_expr = ast.Call(func=self.child_expr,
                              args=[_term_to_ast(a, base_df.dataframe) for a in inputs],
                              keywords=[ast.keyword(arg=k, value=_term_to_ast(v, self))
                                        for k, v in kwargs.items()])
        return DataFrame(expr=child_expr)

    def _test_for_extension(self, name: str):
        'If we have the no-extension flag, then bomb out'
        if _do_not_extend(self):
            raise Exception(f'Object {type(self).__name__} does not have "{name}" defined')

    def __abs__(self):
        '''
        Take the absolute value of ourselves using the python default syntax.
        '''
        self._test_for_extension('abs')
        child_expr = ast.Call(func=ast.Attribute(value=ast_DataFrame(self),
                                                 attr='abs', ctx=ast.Load()),
                              args=[], keywords=[])
        return DataFrame(expr=child_expr)

    def __invert__(self) -> DataFrame:
        ''' Invert, or logical NOT operation. '''
        self._test_for_extension('operator invert')
        child_expr = ast.UnaryOp(op=ast.Invert(), operand=ast_DataFrame(self))
        return DataFrame(child_expr)

    def __and__(self, other) -> Column:
        ''' Bitwise and becomes a logical and. '''
        self._test_for_extension('operator and')
        from .utils import _term_to_ast
        return Column(type(bool), ast.BoolOp(op=ast.And(),
                      values=[_term_to_ast(self, None), _term_to_ast(other, None)]))

    def __or__(self, other) -> Column:
        ''' Bitwise and becomes a logical and. '''
        self._test_for_extension('operator or')
        from .utils import _term_to_ast
        return Column(type(bool), ast.BoolOp(op=ast.Or(),
                      values=[ast.Name('p', ctx=ast.Load()), _term_to_ast(other, self)]))

    def __binary_operator_compare(self, operator: ast.AST, other: Any) -> Column:
        '''Build a column for a binary operation that results in a column of single values.'''

        # How we do this depends on what other is. We need to encode whatever it is in the AST
        # so that it can be properly unpacked.
        from .utils import _term_to_ast
        other_ast = _term_to_ast(other, self)
        compare_ast = ast.Compare(left=_term_to_ast(self, self), ops=[operator],
                                  comparators=[other_ast])
        return Column(type(bool), compare_ast)

    def __binary_operator(self, operator: ast.AST, other: Any) -> DataFrame:
        '''Build a column for a binary operation that results in a column of single values.'''

        # How we do this depends on what other is. We need to encode whatever it is in the AST
        # so that it can be properly unpacked.
        from .utils import _term_to_ast
        other_ast = _term_to_ast(other, self)
        operated = ast.BinOp(left=ast_DataFrame(self), op=operator, right=other_ast)
        return DataFrame(operated)

    def __lt__(self, other) -> Column:
        ''' x < y '''
        self._test_for_extension('operator lt')
        return self.__binary_operator_compare(ast.Lt(), other)

    def __le__(self, other) -> Column:
        ''' x < y '''
        self._test_for_extension('operator le')
        return self.__binary_operator_compare(ast.LtE(), other)

    def __eq__(self, other) -> Column:
        ''' x < y '''
        self._test_for_extension('operator eq')
        return self.__binary_operator_compare(ast.Eq(), other)

    def __ne__(self, other) -> Column:
        ''' x < y '''
        self._test_for_extension('operator ne')
        return self.__binary_operator_compare(ast.NotEq(), other)

    def __gt__(self, other) -> Column:
        ''' x < y '''
        self._test_for_extension('operator gt')
        return self.__binary_operator_compare(ast.Gt(), other)

    def __ge__(self, other) -> Column:
        ''' x < y '''
        self._test_for_extension('operator ge')
        return self.__binary_operator_compare(ast.GtE(), other)

    def __truediv__(self, other) -> DataFrame:
        self._test_for_extension('operator truediv')
        return self.__binary_operator(ast.Div(), other)

    def __mul__(self, other) -> DataFrame:
        self._test_for_extension('operator mul')
        return self.__binary_operator(ast.Mult(), other)

    def __add__(self, other) -> DataFrame:
        self._test_for_extension('operator add')
        return self.__binary_operator(ast.Add(), other)

    def __sub__(self, other) -> DataFrame:
        self._test_for_extension('operator sub')
        return self.__binary_operator(ast.Sub(), other)



import ast
from typing import Callable, Union, Optional, TypeVar
import inspect

from dataframe_expressions import (
    Column, DataFrame, ast_Callable, ast_Column, ast_DataFrame, ast_FunctionPlaceholder)


class DataFrameTypeError(Exception):
    '''Thrown when we don't understand the type in an expression'''
    def __init__(self, message):
        Exception.__init__(self, message)


def _term_to_ast(term: Union[int, str, DataFrame, Column, Callable],
                 parent_df: Optional[Union[DataFrame, Column]]) \
        -> ast.AST:
    '''Return an AST that represents the current term

    Args:
        term        The term (int, string, float, DataFrame, Column, etc.)

    Returns
    '''
    other_ast = None
    if isinstance(term, int) or isinstance(term, float):
        other_ast = ast.Num(n=term)
    elif isinstance(term, str):
        other_ast = ast.Str(s=term)
    elif isinstance(term, DataFrame):
        other_ast = ast_DataFrame(term)
    elif isinstance(term, Column):
        other_ast = ast_Column(term)
    elif callable(term):
        assert parent_df is not None, \
            'Internal Error: Parent DF is required when creating a Callable'
        assert isinstance(parent_df, DataFrame)
        other_ast = ast_Callable(term, parent_df)
    else:
        raise DataFrameTypeError("Do not know how to render a term "
                                 f"of type '{type(term).__name__}'.")

    return other_ast


def user_func(f: Callable) -> Callable:
    '''
    This will allow a function to be embedded into the DataFrame call sequence. For example,

        ```
        @user_func
        def add_it(p1: float) -> float:
            assert False, 'should never be called'
        ```

        And then if you've got `df` defined as your dataframe, you can write:

        ```
        add_it(df.jets.pt)
        ```

        And the resulting `DataFrame` will effectively call `add_it` on each value of `pt` in the
        sequence.

        There are a lot of limitations in this prototype!
    '''
    def emulate_function_call_in_DF(*args):
        f_sig = inspect.signature(f)
        if len(f_sig.parameters) != len(args):
            raise Exception(f'Function {f.__name__} was called with {len(args)} arguments '
                            f'- but needs {len(f_sig.parameters)}')
        f_args = [_term_to_ast(a, None) for a in args]
        call = ast.Call(func=ast_FunctionPlaceholder(f), args=f_args)
        return DataFrame(expr=call)

    return emulate_function_call_in_DF


T = TypeVar('T')


def exclusive_class(o_class: T) -> T:
    '''
    A class that will extend the object model can only access properties that
    are explicitly defined.
    '''
    orig_init = o_class.__init__

    def __init__(self, *args, **kws):
        # The `__no_arb_attr` is a magic string and appears elsewhere in the code
        # as a flag (its value does not matter)
        self.__no_arb_attr = True
        orig_init(self, *args, **kws)  # Call the original __init__

    o_class.__init__ = __init__  # Set the class' __init__ to the new one
    return o_class



import ast


class CloningNodeTransformer(ast.NodeVisitor):
    """
    A :class:`NodeVisitor` subclass that walks the abstract syntax tree and
    allows modification of nodes.
    The `NodeTransformer` will walk the AST and use the return value of the
    visitor methods to replace or remove the old node.  If the return value of
    the visitor method is ``None``, the node will be removed from its location,
    otherwise it is replaced with the return value.  The return value may be the
    original node in which case no replacement takes place.
    Here is an example transformer that rewrites all occurrences of name lookups
    (``foo``) to ``data['foo']``::
       class RewriteName(NodeTransformer):
           def visit_Name(self, node):
               return Subscript(
                   value=Name(id='data', ctx=Load()),
                   slice=Constant(value=node.id),
                   ctx=node.ctx
               )
    Keep in mind that if the node you're operating on has child nodes you must
    either transform the child nodes yourself or call the :meth:`generic_visit`
    method for the node first.
    For nodes that were part of a collection of statements (that applies to all
    statement nodes), the visitor may also return a list of nodes rather than
    just a single node.
    Usually you use the transformer like this::
       node = YourTransformer().visit(node)
    """

    def _get_new_value(self, old_value):
        changed = False
        if isinstance(old_value, list):
            new_values = []
            for value in old_value:
                if isinstance(value, ast.AST):
                    old = value
                    value = self.visit(value)
                    if old is not value:
                        changed = True
                    if value is None:
                        continue
                    elif not isinstance(value, ast.AST):
                        new_values.extend(value)
                        continue
                new_values.append(value)
            return new_values, changed
        elif isinstance(old_value, ast.AST):
            new_node = self.visit(old_value)
            return new_node, new_node is not old_value
            # if new_node is None:
            #     delattr(node, field)
            # else:
            #     setattr(node, field, new_node)
        return old_value, False

    def generic_visit(self, node):
        # Get new and old values
        r = [(field, old_value, self._get_new_value(old_value))
             for field, old_value in ast.iter_fields(node)]

        if len(r) == 0:
            return node

        if all(not i[2][1] for i in r):
            return node

        # Ok - there was a modification. We need to clone the class and pass that
        # back up.

        new_node = node.__class__()
        for f in r:
            if f[1] is not None:
                setattr(new_node, f[0], f[2][0])
        return new_node
